#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#ifdef TURBOC
#include <dos.h>
#include <alloc.h>
#include <graphics.h>
#endif
#include "wirefram.h"

#ifndef PI
#define PI 3.141592653689793
#endif

void MZero (result)
   MATRIX *result;
   {
/* Initialize the matrix to the following values:
   0.0   0.0   0.0   0.0
   0.0   0.0   0.0   0.0
   0.0   0.0   0.0   0.0
   0.0   0.0   0.0   0.0
*/
   register int i, j;

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++)
         (*result)[i][j] = 0.0;
   }

void MIdentity (result)
   MATRIX *result;
   {
/* Initialize the matrix to the following values:
   1.0   0.0   0.0   0.0
   0.0   1.0   0.0   0.0
   0.0   0.0   1.0   0.0
   0.0   0.0   0.0   1.0
*/
   register int i, j;

   for (i = 0 ; i < 4 ; i++)
     for (j = 0 ; j < 4 ; j++)
        if (i == j)
           (*result)[i][j] = 1.0;
        else
           (*result)[i][j] = 0.0;
   }

void MTimes (result, matrix1, matrix2)
   MATRIX *result, *matrix1, *matrix2;
   {
   register int i, j, k;
   MATRIX temp_matrix;

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++) {
         temp_matrix[i][j] = 0.0;
         for (k = 0 ; k < 4 ; k++)
            temp_matrix[i][j] += (*matrix1)[i][k] * (*matrix2)[k][j];
         }

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++)
         (*result)[i][j] = temp_matrix[i][j];
   }

/*  AAC - These are not used, so they are commented out to save code space...

void MAdd (result, matrix1, matrix2)
   MATRIX *result, *matrix1, *matrix2;
   {
   register int i, j;

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++)
         (*result)[i][j] = (*matrix1)[i][j] + (*matrix2)[i][j];
   }

void MSub (result, matrix1, matrix2)
   MATRIX *result, *matrix1, *matrix2;
   {
   register int i, j;

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++)
         (*result)[i][j] = (*matrix1)[i][j] - (*matrix2)[i][j];
   }

void MScale (result, matrix1, amount)
MATRIX *result, *matrix1;
DBL amount;
{
   register int i, j;

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++)
         if (amount == 1.0)
            (*result)[i][j] = (*matrix1)[i][j]; * just copy *
         else
            (*result)[i][j] = (*matrix1)[i][j] * amount;
   return;
}
... up to here! */

void MTranspose (result, matrix1)
   MATRIX *result, *matrix1;
   {
   register int i, j;
   MATRIX temp_matrix;

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++)
         temp_matrix[i][j] = (*matrix1)[j][i];

   for (i = 0 ; i < 4 ; i++)
      for (j = 0 ; j < 4 ; j++)
         (*result)[i][j] = temp_matrix[i][j];
   }


void MTransformVector (result, vector, transformation)
   VECTOR *result, *vector;
   TRANSFORMATION *transformation;
   {
   register int i;
   MATRIX *matrix;
   VECTOR answer;

   matrix = (MATRIX *) transformation -> matrix;

   answer.x = vector->x * (*matrix)[0][0] +
              vector->y * (*matrix)[1][0] +
              vector->z * (*matrix)[2][0] +
              (*matrix)[3][0];
   answer.y = vector->x * (*matrix)[0][1] +
              vector->y * (*matrix)[1][1] +
              vector->z * (*matrix)[2][1] +
              (*matrix)[3][1];
   answer.z = vector->x * (*matrix)[0][2] +
              vector->y * (*matrix)[1][2] +
              vector->z * (*matrix)[2][2] +
              (*matrix)[3][2];

   result->x = answer.x;
   result->y = answer.y;
   result->z = answer.z;
   }

void MTransVector (result, vector, transformation)
   VECTOR *result, *vector;
   TRANSFORMATION *transformation;
   {
   register int i;
   DBL answer_array[4];
   MATRIX *matrix;

   matrix = (MATRIX *) transformation -> matrix;

   for (i = 0 ; i < 3 ; i++)
      answer_array[i] = vector -> x * (*matrix)[0][i]
                      + vector -> y * (*matrix)[1][i]
                      + vector -> z * (*matrix)[2][i];

   result -> x  = answer_array[0];
   result -> y  = answer_array[1];
   result -> z  = answer_array[2];
   }

void MInverseTransformVector (result, vector, transformation)
   VECTOR *result, *vector;
   TRANSFORMATION *transformation;
   {
   register int i;
   DBL answer_array[4];
   MATRIX *matrix;

   matrix = (MATRIX *) transformation -> inverse;

   for (i = 0 ; i < 4 ; i++)
      answer_array[i] = vector -> x * (*matrix)[0][i]
                      + vector -> y * (*matrix)[1][i]
                      + vector -> z * (*matrix)[2][i]
                      + (*matrix)[3][i];

   result -> x  = answer_array[0];
   result -> y  = answer_array[1];
   result -> z  = answer_array[2];
   }

void MInverseTransVector (result, vector, transformation)
   VECTOR *result, *vector;
   TRANSFORMATION *transformation;
   {
   register int i;
   DBL answer_array[4];
   MATRIX *matrix;

   matrix = (MATRIX *) transformation -> inverse;

   for (i = 0 ; i < 4 ; i++)
      answer_array[i] = vector -> x * (*matrix)[0][i]
                      + vector -> y * (*matrix)[1][i]
                      + vector -> z * (*matrix)[2][i];

   result -> x  = answer_array[0];
   result -> y  = answer_array[1];
   result -> z  = answer_array[2];
   }

void Get_Scaling_Transformation (result, vector)
   TRANSFORMATION *result;
   VECTOR *vector;
   {
   MIdentity(&result->matrix);
   (result->matrix)[0][0] = vector->x;
   (result->matrix)[1][1] = vector->y;
   (result->matrix)[2][2] = vector->z;

   MIdentity(&result->inverse);
   (result -> inverse)[0][0] = 1.0 / vector -> x;
   (result -> inverse)[1][1]= 1.0 / vector -> y;
   (result -> inverse)[2][2] = 1.0 / vector -> z;
   }

void Get_Translation_Transformation (transformation, vector)
   TRANSFORMATION *transformation;
   VECTOR *vector;
   {
   MIdentity((MATRIX *)&transformation->matrix);
   (transformation->matrix)[3][0] = vector->x;
   (transformation->matrix)[3][1] = vector->y;
   (transformation->matrix)[3][2] = vector->z;

   MIdentity ((MATRIX *)&transformation->inverse);
   (transformation->inverse)[3][0] = 0.0 - vector->x;
   (transformation->inverse)[3][1] = 0.0 - vector->y;
   (transformation->inverse)[3][2] = 0.0 - vector->z;
   }

void Get_Rotation_Transformation (transformation, vector)
   TRANSFORMATION *transformation;
   VECTOR *vector;
   {
   MATRIX Matrix;
   VECTOR Radian_Vector;
   register DBL cosx, cosy, cosz, sinx, siny, sinz;

   VScale (Radian_Vector, *vector, PI/180.0);
   MIdentity(&transformation->matrix);
   cosx = cos (Radian_Vector.x);
   sinx = sin (Radian_Vector.x);
   cosy = cos (Radian_Vector.y);
   siny = sin (Radian_Vector.y);
   cosz = cos (Radian_Vector.z);
   sinz = sin (Radian_Vector.z);

   (transformation->matrix) [1][1] = cosx;
   (transformation->matrix) [2][2] = cosx;
   (transformation->matrix) [1][2] = sinx;
   (transformation->matrix) [2][1] = 0.0 - sinx;
   MTranspose(&transformation->inverse, &transformation->matrix);

   MIdentity(&Matrix);
   Matrix[0][0] = cosy;
   Matrix[2][2] = cosy;
   Matrix[0][2] = 0.0 - siny;
   Matrix[2][0] = siny;
   MTimes(&transformation->matrix, &transformation->matrix, &Matrix);
   MTranspose(&Matrix, &Matrix);
   MTimes(&transformation->inverse, &Matrix, &transformation->inverse);

   MIdentity(&Matrix);
   Matrix [0][0] = cosz;
   Matrix [1][1] = cosz;
   Matrix [0][1] = sinz;
   Matrix [1][0] = 0.0 - sinz;
   MTimes(&transformation->matrix, &transformation->matrix, &Matrix);
   MTranspose(&Matrix, &Matrix);
   MTimes(&transformation->inverse, &Matrix, &transformation->inverse);
   }

void Compose_Transformations (Original_Transformation, New_Transformation)
   TRANSFORMATION *Original_Transformation, *New_Transformation;
   {
   MTimes(&Original_Transformation->matrix,
          &Original_Transformation->matrix,
          &New_Transformation->matrix);

   MTimes(&Original_Transformation->inverse,
          &New_Transformation->inverse,
          &Original_Transformation->inverse);
   }

void
Get_Perspective_Transformation(TRANSFORMATION *transformation, DBL dist)
{
   MIdentity(&transformation->matrix);
   transformation->matrix[2][3] = 1.0 / dist;
   MIdentity(&transformation->inverse);
}

TRANSFORMATION *
Get_Transformation()
{
  TRANSFORMATION *New_Transformation;
  if ((New_Transformation =
        (TRANSFORMATION *) malloc (sizeof (TRANSFORMATION))) == NULL) {
    printf("Cannot allocate transformation");
    exit(1);
    }
  MIdentity(&New_Transformation->matrix);
  MIdentity(&New_Transformation->inverse);
  return(New_Transformation);
}
